#include <cmath>
#include <iostream>

#include <Eigen/Dense>

Eigen::Vector2d f(const Eigen::Vector2d &x) {
	return Eigen::Vector2d(x(1), 2.0 * x(0) * (1.0 + x(0) * x(0)));
}

Eigen::Vector2d F(const Eigen::Vector2d &x, const Eigen::Vector2d &zk, double h) {
	return zk + h * f(x) - x;
}

Eigen::Matrix2d DF(const Eigen::Vector2d &x, double h) {
	double d = h * (6.0 * x(0) * x(0) + 2.0);
	Eigen::Matrix2d M;
	M << -1.0, h, d, -1.0;
	return M;
}

Eigen::Vector2d Newton(Eigen::Vector2d x, double h, int n, double tol) {
	const Eigen::Vector2d zk = x;
	for (int i = 0; i < n; ++i) {
		if (F(x, zk, h).squaredNorm() <= tol * tol) return x;
		x += -DF(x, h).fullPivLu().solve(F(x, zk, h));
	}
	std::cout << "Warning: Maximal number of iterations reached." << std::endl;
	return x;
}

Eigen::Vector2d QuasiNewton(Eigen::Vector2d x, double h, int n, double tol) {
	const Eigen::Vector2d zk = x;
	Eigen::Matrix2d J_inv = DF(x, h).inverse();
	for (int i = 0; i < n; ++i) {
		Eigen::Vector2d dx = -J_inv * F(x, zk, h);
		x += dx;

		if (F(x, zk, h).squaredNorm() <= tol * tol) return x;

		Eigen::Vector2d T = J_inv * F(x, zk, h);
		J_inv += (-T * dx.transpose() / (dx.squaredNorm() + dx.transpose() * T)) * J_inv;
	}
	std::cout << "Warning: Maximal number of iterations reached." << std::endl;
	return x;
}

Eigen::Vector2d ImplicitEuler(Eigen::Vector2d z, double h, int N) {
	for (int k = 0; k < N; ++k) {
		z = Newton(z, h, 10, 1.0e-8); //z = QuasiNewton(z, h, 10, 1.0e-8);
	}
	return z;
}


int main() {
	int N = 100;
	double h = M_PI / (4 * N);
	Eigen::Vector2d z0(0.0, 1.0);
	std::cout << "endtime =    " << N * h << std::endl;
	std::cout << "Impl. Euler: " << ImplicitEuler(z0, h, N)(0) << std::endl;
	std::cout << "exact:       " << std::tan(N * h) << std::endl;
	return 0;
}
